import greenfoot.*;  // （World、Actor、GreenfootImage、GreenfootSound、Greenfoot及MouseInfo）
import java.awt.Color;
import java.util.List;

/**
 * 球。
 * 
 * @author 依瑪貓
 * @version 2012-08-28
 */
public class Ball extends Actor
{
    /** 球的預設半徑。 */
    private static final int defaultRadius = 10;
    
    /** 球的顏色。 */
    private static final Color color = Color.YELLOW;
    
    /** 球的半徑。 */
    private int radius = defaultRadius;
    
    /** X 軸的速度。 */
    private int speedX = 2;
    
    /** Y 軸的速度。 */
    private int speedY = 5;
    
    /**
     * 建立一顆新球。
     * 
     */
    public Ball() {
        drawImage();
    }
    
    /**
     * 繪製球。
     * 
     */
    public void drawImage() 
    {
        GreenfootImage image = new GreenfootImage(radius * 2 + 1, radius * 2 + 1);
        image.setColor(color);
        image.fillOval(0, 0, image.getWidth(), image.getHeight());
        setImage(image);
    }
    
    /**
     * 每次按下「單步執行」或「執行」按鈕，都會呼叫這個方法。
     * 
     */
    public void act() 
    {
        move();
        bounceAtWorldEdge();
        bounceAtBar();
        checkHitBrick();
        gameOverWhenDropped();
    }
    
    /**
     * 往前移動。
     * 
     */
    private void move() {
        setLocation(getX() + speedX, getY() + speedY);
    }
    
    /**
     * 出界的時候反彈。
     * 
     */
    private void bounceAtWorldEdge() {
        World space = getWorld();
        if (getX() - radius < 0 || getX() + radius >= space.getWidth()) {
            speedX = -speedX;
        }
        if (getY() - radius < 0) {
            speedY = -speedY;
        }
    }
    
    /**
     * 碰到擊球板的時候反彈。
     * 
     */
    private void bounceAtBar() {
        List<Bar> bars = getWorld().getObjects(Bar.class);
        Bar bar = bars.get(0);
        if (isTouchedFromUpBottom(bar)) {
            speedY = Math.abs(speedY);
            if (getY() < bar.getY()) {
                speedY = -speedY;
            }
        } else if (isTouchedFromLeftRight(bar)) {
            speedX = Math.abs(speedX);
            if (getX() < bar.getX()) {
                speedX = -speedX;
            }
        } else if (isTouchedCorner(bar)) {
            speedX = -speedX;
            speedY = -speedY;
        }
    }
    
    /**
     * 檢查有沒有碰到磚塊。
     * 
     */
    private void checkHitBrick() {
        Space space = (Space) getWorld();
        List<Brick> bricks = space.getObjects(Brick.class);
        Score score = space.getScore();
        for (Brick brick : bricks) {
            if (isTouchedFromUpBottom(brick)) {
                speedY = Math.abs(speedY);
                if (getY() < brick.getY()) {
                    speedY = -speedY;
                }
                space.removeObject(brick);
                score.increase();
            } else if (isTouchedFromLeftRight(brick)) {
                speedX = Math.abs(speedX);
                if (getX() < brick.getX()) {
                    speedX = -speedX;
                }
                space.removeObject(brick);
                score.increase();
            } else if (isTouchedCorner(brick)) {
                speedX = -speedX;
                speedY = -speedY;
                space.removeObject(brick);
                score.increase();
            }
        }
    }
    
    /**
     * 有沒有從上下碰撞到長方形。
     * 
     * @param rect 要檢查碰撞的長方形。
     * @return 有碰撞到時傳回 true ，否則傳回 false 。
     */
    private boolean isTouchedFromUpBottom(Rect rect) {
        if (    Math.abs(rect.getY() - getY()) <= rect.getRadiusY() + radius
                && getX() >= rect.getX() - rect.getRadiusX()
                && getX() < rect.getX() + rect.getRadiusX()) {
            return true;
        }
        return false;
    }
    
    /**
     * 有沒有從左右碰撞到長方形。
     * 
     * @param rect 要檢查碰撞的長方形。
     * @return 有碰撞到時傳回 true ，否則傳回 false 。
     */
    private boolean isTouchedFromLeftRight(Rect rect) {
        if (    Math.abs(rect.getX() - getX()) <= rect.getRadiusX() + radius
                && getY() >= rect.getY() - rect.getRadiusY()
                && getY() < rect.getY() + rect.getRadiusY()) {
            return true;
        }
        return false;
    }
    
    /**
     * 有沒有從邊角碰撞到長方形的邊角。
     * 
     * @param rect 要檢查碰撞的長方形。
     * @return 有碰撞到時傳回 true ，否則傳回 false 。
     */
    private boolean isTouchedCorner(Rect rect) {
        int cornerX = -1, cornerY = -1, dX = 0, dY = 0;
        
        // 左上角。
        cornerX = rect.getX() - rect.getRadiusX();
        dX = getX() - cornerX;
        cornerY = rect.getY() - rect.getRadiusY();
        dY = getY() - cornerY;
        if (dX < 0 && dY < 0 && dX * dX + dY * dY <= radius * radius) {
            return true;
        }
        
        // 右上角。
        cornerX = rect.getX() + rect.getRadiusX();
        dX = getX() - cornerX;
        cornerY = rect.getY() - rect.getRadiusY();
        dY = getY() - cornerY;
        if (dX > 0 && dY < 0 && dX * dX + dY * dY <= radius * radius) {
            return true;
        }
        
        // 左下角。
        cornerX = rect.getX() - rect.getRadiusX();
        dX = getX() - cornerX;
        cornerY = rect.getY() + rect.getRadiusY();
        dY = getY() - cornerY;
        if (dX < 0 && dY > 0 && dX * dX + dY * dY <= radius * radius) {
            return true;
        }
        
        // 右上角。
        cornerX = rect.getX() + rect.getRadiusX();
        dX = getX() - cornerX;
        cornerY = rect.getY() + rect.getRadiusY();
        dY = getY() - cornerY;
        if (dX > 0 && dY > 0 && dX * dX + dY * dY <= radius * radius) {
            return true;
        }
        
        return false;
    }
    
    /**
     * 球掉下來的時候遊戲結束。
     * 
     */
    private void gameOverWhenDropped() {
        Space space = (Space) getWorld();
        if (getY() + radius >= space.getHeight()) {
            space.gameOver();
        }
    }
}
